#include <Wire.h>
#include <WiiSensors.h>
#include <Kalman.h>
#include <PID_v1.h>
#include <SoftwareSerial.h>

SoftwareSerial SerialB(16, 14);

WiiSensors wiiSensors;

Kalman kalmanYaw;
Kalman kalmanPitch;
Kalman kalmanRoll;

int cnt1 = 0, cnt2 = 0, whichVar = 1, whichParam = 1;

double accXangle, accYangle, accZangle;
double YawRate, PitchRate, RollRate, YawAngle;
double Yaw = 0, Pitch = 0, Roll = 0;     //Valores de inclinacion angular (Entradas de los PID)
double OutYaw, OutPitch, OutRoll, SetYaw = -1, SetPitch = 144.40, SetRoll = 145.15;  //Valores de Setpoint neutros y de Output para los tres PID
PID PIDYaw(&Yaw, &OutYaw, &SetYaw, 0.1, 0, 0, DIRECT);
PID PIDPitch(&Pitch, &OutPitch, &SetPitch, 0.7, 0, 0.1, DIRECT);
PID PIDRoll(&Roll, &OutRoll, &SetRoll, 0.7, 0, 0.1, DIRECT);

uint32_t timer;

#define motorFL  5
#define motorFR  6
#define motorBL  9
#define motorBR  10

int valor;
boolean armed = false, comm = false, tunningPID = false;
double throttle = 80, throttleFL, throttleFR, throttleBL, throttleBR;

void setup()
{
  SerialB.begin(9600);
  pinMode(4, OUTPUT);    //Alimenta a los sensores de Wii
  digitalWrite(4, HIGH);
  pinMode(7, OUTPUT);    //Pin para switchToWmp()
  digitalWrite(7, HIGH);
  pinMode(8, OUTPUT);    //Pin para switchToNunchuck()
  digitalWrite(8, HIGH);
  delay(310);            //Hay que esperar 310 ms para calibrar el WMP despues de encenderlo

  analogWrite(motorFL, 80);
  analogWrite(motorFR, 80);
  analogWrite(motorBL, 80);
  analogWrite(motorBR, 80);

  SerialB.println("Iniciando...");
  PIDYaw.SetMode(AUTOMATIC);
  PIDYaw.SetSampleTime(50);
  PIDYaw.SetOutputLimits(-50, 50);
  PIDPitch.SetMode(AUTOMATIC);
  PIDPitch.SetSampleTime(50);
  PIDPitch.SetOutputLimits(-50, 50);
  PIDRoll.SetMode(AUTOMATIC);
  PIDRoll.SetSampleTime(50);
  PIDRoll.SetOutputLimits(-50, 50);

  Wire.begin();        // Inicializa la comunicacion en la direccion 0x52
  
  CalibrarSensores();
}

void loop()
{
  LeerControles();
  if(comm == false)
  {
    cnt2++;
    if(cnt2 >= 100)
    {
      armed = false;
      throttle = 80;
      analogWrite(motorFL, 80);
      analogWrite(motorFR, 80);
      analogWrite(motorBL, 80);
      analogWrite(motorBR, 80);
      SerialB.println("Se ha perdido la conexion");
      while(!comm)
      {
        LeerControles();
      }
      cnt2 = 0;
    }
  }
  else
  {
    cnt2 = 0;
  }

  LeerSensores();

  PIDYaw.Compute();
  PIDPitch.Compute();
  PIDRoll.Compute();

  throttleFL = throttle - OutPitch + OutRoll + OutYaw;
  throttleFR = throttle - OutPitch - OutRoll - OutYaw;
  throttleBL = throttle + OutPitch + OutRoll - OutYaw;
  throttleBR = throttle + OutPitch - OutRoll + OutYaw;
  if(throttle > 100)
  {
    throttleFL = constrain(throttleFL, 100, 254);
    throttleFR = constrain(throttleFR, 100, 254);
    throttleBL = constrain(throttleBL, 100, 254);
    throttleBR = constrain(throttleBR, 100, 254);
    analogWrite(motorFL, throttleFL);
    analogWrite(motorFR, throttleFR);
    analogWrite(motorBL, throttleBL);
    analogWrite(motorBR, throttleBR);
  }
  else
  {
    analogWrite(motorFL, throttle);
    analogWrite(motorFR, throttle);
    analogWrite(motorBL, throttle);
    analogWrite(motorBR, throttle);
  }
  
  Mensajes();
  
  cnt1++;
  comm = false;
  delay(1);
}

void LeerSensores()
{
  wiiSensors.switchToWmp();
  wiiSensors.receiveData();
  wiiSensors.switchToNunchuck();
  wiiSensors.receiveNunchuckData();
  //accZangle = (atan2(-wiiSensors.accel_x_axis, -wiiSensors.accel_y_axis) + PI) * RAD_TO_DEG;
  accXangle = (atan2(-wiiSensors.accel_y_axis, wiiSensors.accel_z_axis) + PI) * RAD_TO_DEG;
  accYangle = (atan2(-wiiSensors.accel_x_axis, wiiSensors.accel_z_axis) + PI) * RAD_TO_DEG;
  YawRate = wiiSensors.yaw / 131.0;
  PitchRate = wiiSensors.pitch / 131.0;
  RollRate = wiiSensors.roll / 131.0;
  Yaw = kalmanYaw.getAngle(wiiSensors.yaw, YawRate, (double)(micros() - timer) / 1000000);
  Pitch = kalmanPitch.getAngle(accXangle, PitchRate, (double)(micros() - timer) / 1000000);
  Roll = kalmanRoll.getAngle(accYangle, RollRate, (double)(micros() - timer) / 1000000);
  timer = micros();
}

void CalibrarSensores()
{
  wiiSensors.switchToWmp();
  wiiSensors.wmpOn(); //turn WM+ on
  SerialB.println("Calibrando...");
  delay(100);
  wiiSensors.calibrateZeroes(); //calibra los ceros del WMP
  wiiSensors.switchToNunchuck();
  wiiSensors.nunchuck_init (); // send the initilization handshake

  //accZangle = (atan2(-wiiSensors.accel_x_axis, -wiiSensors.accel_y_axis) + PI) * RAD_TO_DEG;
  accXangle = (atan2(-wiiSensors.accel_y_axis, wiiSensors.accel_z_axis) + PI) * RAD_TO_DEG;
  accYangle = (atan2(-wiiSensors.accel_x_axis, wiiSensors.accel_z_axis) + PI) * RAD_TO_DEG;
  kalmanYaw.setAngle(0);  //Establece los angulos para el filtro de Kalman
  kalmanPitch.setAngle(accXangle);
  kalmanRoll.setAngle(accYangle);
  for(int i = 0; i < 1500; i++)    //Lee 1500 veces para conseguir un valor estable
  {
    LeerSensores();
  }
  SetYaw = Yaw;      //Almacena los valores neutros en los Setpoints
  SetPitch = Pitch;
  SetRoll = Roll;
  SerialB.println("Calibracion terminada");
  timer = micros();
}

void LeerControles()
{
  if(SerialB.available() > 0)
  {
    valor = SerialB.read() - 48;
    switch(valor)
    {
    case 2:  //T+
      if(armed)
      {
        throttle++;
        SerialB.print("Throttle: ");
        SerialB.println(throttle);
      }
      break;
    case 4:  //Y-
      if(armed)
      {
        SetYaw--;
        SerialB.print("SetYaw: ");
        SerialB.println(SetYaw);
      }
      break;
    case 5:  //ON/OFF
      if(armed)
      {
        armed = false;
        throttle = 80;
        analogWrite(motorFL, 80);
        analogWrite(motorFR, 80);
        analogWrite(motorBL, 80);
        analogWrite(motorBR, 80);
        SerialB.println("Dron desactivado");
        delay(2000);              //Espera a que los motores se paren
        digitalWrite(4, LOW);    //Apaga los sensores
        delay(100);              //Espera 0.1 seg
        digitalWrite(4, HIGH);    //Reactiva los sensores
        delay(310);
        CalibrarSensores();
        PIDYaw.SetMode(AUTOMATIC);  //Reinicia los PIDs
        PIDPitch.SetMode(AUTOMATIC);
        PIDRoll.SetMode(AUTOMATIC);
      }
      else
      {
        throttle = 100;
        armed = true;
        SerialB.println("Dron activado");
      }
      break;
    case 6:  //Y+
      if(armed)
      {
        SetYaw++;
        SerialB.print("SetYaw: ");
        SerialB.println(SetYaw);
      }
      break;
    case 8:  //T-
      if(armed && throttle > 100)
      {
        throttle--;
        SerialB.print("Throttle: ");
        SerialB.println(throttle);
      }
      break;
    case 49:  //P+
      if(armed)
      {
        SetPitch++;
        SerialB.print("SetPitch: ");
        SerialB.println(SetPitch);
      }
      break;  //R-
    case 50:
      if(armed)
      {
        SetRoll--;
        SerialB.print("SetRoll: ");
        SerialB.println(SetRoll);
      }
      break;
    case 51:  //PID
      if(armed)
      {
        //SerialB.println("Desconecta primero el dron");
        throttle = 160;
      }
      else
      {
        SerialB.println("Cambia los valores PID:");
        TunePID();
      }
      break;
    case 52:  //R+
      if(armed)
      {
        SetRoll++;
        SerialB.print("SetRoll: ");
        SerialB.println(SetRoll);
      }
      break;
    case 53:  //P-
      if(armed)
      {
        SetPitch--;
        SerialB.print("SetPitch: ");
        SerialB.println(SetPitch);
      }
      break;
    case 58:
      comm = true;
      break;
    default:
      break;
    }
  }
}

void TunePID()
{
  tunningPID = true;
  SerialB.println("(Incrementos de 0.1)");
  SerialB.println("(Parametro Kp)");
  while(tunningPID == true)
  {
    if(SerialB.available() > 0)
    {
      valor = SerialB.read() - 48;
      switch(valor)
      {
      case 2:  //T+ PID
        if(whichParam == 1)
        {
          if(whichVar == 1)
          {
            PIDYaw.SetTunings(PIDYaw.GetKp() + 0.01, PIDYaw.GetKi(), PIDYaw.GetKd());
            SerialB.print("Kp_Yaw = ");
            SerialB.println(PIDYaw.GetKp());
          }
          if(whichVar == 2)
          {
            PIDPitch.SetTunings(PIDPitch.GetKp() + 0.01, PIDPitch.GetKi(), PIDPitch.GetKd());
            SerialB.print("Kp_Pitch = ");
            SerialB.println(PIDPitch.GetKp());
          }
          if(whichVar == 3)
          {
            PIDRoll.SetTunings(PIDRoll.GetKp() + 0.01, PIDRoll.GetKi(), PIDRoll.GetKd());
            SerialB.print("Kp_Roll = ");
            SerialB.println(PIDRoll.GetKp());
          }
        }
        if(whichParam == 2)
        {
          if(whichVar == 1)
          {
            PIDYaw.SetTunings(PIDYaw.GetKp(), PIDYaw.GetKi() + 0.01, PIDYaw.GetKd());
            SerialB.print("Ki_Yaw = ");
            SerialB.println(PIDYaw.GetKi());
          }
          if(whichVar == 2)
          {
            PIDPitch.SetTunings(PIDPitch.GetKp(), PIDPitch.GetKi() + 0.01, PIDPitch.GetKd());
            SerialB.print("Ki_Pitch = ");
            SerialB.println(PIDPitch.GetKi());
          }
          if(whichVar == 3)
          {
            PIDRoll.SetTunings(PIDRoll.GetKp(), PIDRoll.GetKi() + 0.01, PIDRoll.GetKd());
            SerialB.print("Ki_Roll = ");
            SerialB.println(PIDRoll.GetKi());
          }
        }
        if(whichParam == 3)
        {
          if(whichVar == 1)
          {
            PIDYaw.SetTunings(PIDYaw.GetKp(), PIDYaw.GetKi(), PIDYaw.GetKd() + 0.01);
            SerialB.print("Kd_Yaw = ");
            SerialB.println(PIDYaw.GetKd());
          }
          if(whichVar == 2)
          {
            PIDPitch.SetTunings(PIDPitch.GetKp(), PIDPitch.GetKi(), PIDPitch.GetKd() + 0.01);
            SerialB.print("Kd_Pitch = ");
            SerialB.println(PIDPitch.GetKd());
          }
          if(whichVar == 3)
          {
            PIDRoll.SetTunings(PIDRoll.GetKp(), PIDRoll.GetKi(), PIDRoll.GetKd() + 0.01);
            SerialB.print("Kd_Roll = ");
            SerialB.println(PIDRoll.GetKd());
          }
        }
        break;
      case 6:  //Y PID
        whichVar = 1;
        SerialB.println("Yaw: ");

        break;
      case 8:  //T- PID
        if(whichParam == 1)
        {
          if(whichVar == 1)
          {
            PIDYaw.SetTunings(PIDYaw.GetKp() - 0.01, PIDYaw.GetKi(), PIDYaw.GetKd());
            SerialB.print("Kp_Yaw = ");
            SerialB.println(PIDYaw.GetKp());
          }
          if(whichVar == 2)
          {
            PIDPitch.SetTunings(PIDPitch.GetKp() - 0.01, PIDPitch.GetKi(), PIDPitch.GetKd());
            SerialB.print("Kp_Pitch = ");
            SerialB.println(PIDPitch.GetKp());
          }
          if(whichVar == 3)
          {
            PIDRoll.SetTunings(PIDRoll.GetKp() - 0.01, PIDRoll.GetKi(), PIDRoll.GetKd());
            SerialB.print("Kp_Roll = ");
            SerialB.println(PIDRoll.GetKp());
          }
        }
        if(whichParam == 2)
        {
          if(whichVar == 1)
          {
            PIDYaw.SetTunings(PIDYaw.GetKp(), PIDYaw.GetKi() - 0.01, PIDYaw.GetKd());
            SerialB.print("Ki_Yaw = ");
            SerialB.println(PIDYaw.GetKi());
          }
          if(whichVar == 2)
          {
            PIDPitch.SetTunings(PIDPitch.GetKp(), PIDPitch.GetKi() - 0.01, PIDPitch.GetKd());
            SerialB.print("Ki_Pitch = ");
            SerialB.println(PIDPitch.GetKi());
          }
          if(whichVar == 3)
          {
            PIDRoll.SetTunings(PIDRoll.GetKp(), PIDRoll.GetKi() - 0.01, PIDRoll.GetKd());
            SerialB.print("Ki_Roll = ");
            SerialB.println(PIDRoll.GetKi());
          }
        }
        if(whichParam == 3)
        {
          if(whichVar == 1)
          {
            PIDYaw.SetTunings(PIDYaw.GetKp(), PIDYaw.GetKi(), PIDYaw.GetKd() - 0.01);
            SerialB.print("Kd_Yaw = ");
            SerialB.println(PIDYaw.GetKd());
          }
          if(whichVar == 2)
          {
            PIDPitch.SetTunings(PIDPitch.GetKp(), PIDPitch.GetKi(), PIDPitch.GetKd() - 0.01);
            SerialB.print("Kd_Pitch = ");
            SerialB.println(PIDPitch.GetKd());
          }
          if(whichVar == 3)
          {
            PIDRoll.SetTunings(PIDRoll.GetKp(), PIDRoll.GetKi(), PIDRoll.GetKd() - 0.01);
            SerialB.print("Kd_Roll = ");
            SerialB.println(PIDRoll.GetKd());
          }
        }
        break;
      case 49:  //P PID
        whichVar = 2;
        SerialB.println("Pitch: ");

        break;
      case 51:
        whichParam++;
        if(whichParam == 2)
        {
          SerialB.println("Parametro Ki");
        }
        if(whichParam == 3)
        {
          SerialB.println("Parametro Kd");
        }
        if(whichParam == 4)
        {
          tunningPID = false;
          SerialB.println("Conf PID terminada");
        }
        break;
      case 52:  //R PID
        whichVar = 3;
        SerialB.println("Roll: ");

        break;
      case 58:
        comm = true;
        break;
      default:
        SerialB.println("Opcion incorrecta");
        break;
      }
    }
  }
  whichParam = 1;
}

void Mensajes()
{
  if(cnt1 == 100)  //Salidas por pantalla
  {
    /*SerialB.print("accel_x_axis: ");
    SerialB.print(wiiSensors.accel_x_axis);
    SerialB.print("\t\taccel_y_axis: ");
    SerialB.print(wiiSensors.accel_y_axis);
    SerialB.print("\t\taccel_z_axis: ");
    SerialB.println(wiiSensors.accel_z_axis);*/

    /*SerialB.print("accXangle: ");
    SerialB.print(accXangle);
    SerialB.print("\t\taccYangle: ");
    SerialB.print(accYangle);
    SerialB.print("\t\taccZangle: ");
    SerialB.println(accZangle);*/

    /*SerialB.print("YawAccel: ");
    SerialB.print(wiiSensors.yaw);
    SerialB.print("\t\tPitchAccel: ");
    SerialB.print(wiiSensors.pitch);
    SerialB.print("\t\tRollAccel: ");
    SerialB.println(wiiSensors.roll);*/

    /*SerialB.print("Yaw: ");
    SerialB.print(Yaw);
    SerialB.print("\t\tPitch: ");
    SerialB.print(Pitch);
    SerialB.print("\t\tRoll: ");
    SerialB.println(Roll);*/

    SerialB.print("OutYaw: ");
    SerialB.print(OutYaw);
    SerialB.print("\t\tOutPitch: ");
    SerialB.print(OutPitch);
    SerialB.print("\t\tOutRoll: ");
    SerialB.println(OutRoll);

    /*SerialB.print("throttleFL: ");
    SerialB.print(throttleFL);
    SerialB.print("\t\tthrottleFR: ");
    SerialB.print(throttleFR);
    SerialB.print("\t\tthrottleBL: ");
    SerialB.print(throttleBL);
    SerialB.print("\t\tthrottleBR: ");
    SerialB.println(throttleBR);*/
    
    //SerialB.println(throttle);
    cnt1 = 0;
  }
}
