#if ARDUINO >= 100
  #include "Arduino.h"
#else
  #include "WProgram.h"
#endif
#include <WiiSensors.h>
#include <Wire.h>
#include <string.h>

#undef int
#include <stdio.h>

byte data[6]; //six data bytes
uint8_t outbuf[6];        // array to store arduino output
int cnt = 0;

WiiSensors::WiiSensors()
{

}

void WiiSensors::wmpOn()
{
  Wire.beginTransmission(0x53); //WM+ starts out deactivated at address 0x53
  Wire.write(0xfe); //send 0x04 to address 0xFE to activate WM+
  Wire.write(0x04);
  Wire.endTransmission(); //WM+ jumps to address 0x52 and is now active
}

void WiiSensors::wmpOff()
{
  Wire.beginTransmission(82);
  Wire.write(0xf0);//address then
  Wire.write(0x55);//command
  //Wire.write(0x00);
  //Wire.write(0xfb);
  Wire.endTransmission();
}

void WiiSensors::wmpSendZero()
{
  Wire.beginTransmission(0x52); //now at address 0x52
  Wire.write(0x00); //send zero to signal we want info
  Wire.endTransmission();
}

void WiiSensors::calibrateZeroes()
{
  for (int i = 0; i < 10; i++)
  {
    wmpSendZero();
    Wire.requestFrom(0x52,6);
    for (int i = 0; i < 6; i++)
    {
      data[i] = Wire.read();
    }
    yaw0 += (((data[3] >> 2) << 8) + data[0]) / 10; //average 10 readings
    pitch0 += (((data[4] >> 2) << 8) + data[1]) / 10;
    roll0 += (((data[5]>>2) << 8) + data[2]) / 10;
  }
  Serial.print("Yaw0:");
  Serial.print(yaw0);
  Serial.print(" Pitch0:");
  Serial.print(pitch0);
  Serial.print(" Roll0:");
  Serial.println(roll0);
}

void WiiSensors::receiveData()
{
  wmpSendZero(); //send zero before each request (same as nunchuck)
  Wire.requestFrom(0x52, 6); //request the six bytes from the WM+
  for (int i=0;i < 6;i++)
  {
    data[i] = Wire.read();
  }
  yaw = ((data[3] >> 2) << 8) + data[0] - yaw0;
  pitch = ((data[4] >> 2) << 8) + data[1] - pitch0;
  roll = ((data[5] >> 2) << 8) + data[2] - roll0;
}

void WiiSensors::receiveRaw()
{
  wmpSendZero(); //send zero before each request (same as nunchuck)
  Wire.requestFrom(0x52,6);//request the six bytes from the WM+
  for(int i = 0; i < 6; i++)
  {
    data[i] = Wire.read();
  }
  yaw = ((data[3] >> 2) << 8) + data[0];
  pitch = ((data[4] >> 2) << 8) + data[1];
  roll = ((data[5] >> 2) << 8) + data[2];
}

void WiiSensors::switchToWmp()
{
  digitalWrite(7, LOW);
  digitalWrite(8, LOW);
  digitalWrite(7, HIGH);
}

void WiiSensors::switchToNunchuck()
{
  digitalWrite(7, LOW);
  digitalWrite(8, LOW);
  digitalWrite(8, HIGH);
}

void WiiSensors::nunchuck_init()
{
  Wire.beginTransmission (0x52);    // transmit to device 0x52
  Wire.write (0x40);        // sends memory address
  Wire.write (0x00);        // sends sent a zero. 
  Wire.endTransmission ();    // stop transmitting
}

void WiiSensors::send_zero()
{
  Wire.beginTransmission(0x52);    // transmit to device 0x52
  Wire.write(0x00);        // sends one byte
  Wire.endTransmission();    // stop transmitting
}

void WiiSensors::receiveNunchuckData()
{
  Wire.requestFrom(0x52, 6);
  for (int i = 0; i < 6; i++)
  {
    data[i] = Wire.read();
  }
  make_nunchuck_data();
  send_zero();
}

void WiiSensors::make_nunchuck_data()
{
  for(int i = 0; i < 6; i++)
  {
    outbuf[i] = nunchuck_decode_byte(data[i]);
  }
  int joy_x_axis = outbuf[0];
  int joy_y_axis = outbuf[1];
  accel_x_axis = outbuf[2] * 2 * 2;
  accel_y_axis = outbuf[3] * 2 * 2;
  accel_z_axis = outbuf[4] * 2 * 2;
  boolean z_button = 0;
  boolean c_button = 0;

  // byte outbuf[5] contains bits for z and c buttons
  // it also contains the least significant bits for the accelerometer data
  // so we have to check each bit of byte outbuf[5]
  if((outbuf[5] >> 0) & 1)
  {
    z_button = 1;
  }
  if((outbuf[5] >> 1) & 1)
  {
    c_button = 1;
  }
  if((outbuf[5] >> 2) & 1)
  {
    accel_x_axis += 2;
  }
  if((outbuf[5] >> 3) & 1)
  {
    accel_x_axis += 1;
  }
  if((outbuf[5] >> 4) & 1)
  {
    accel_y_axis += 2;
  }
  if((outbuf[5] >> 5) & 1)
  {
    accel_y_axis += 1;
  }
  if((outbuf[5] >> 6) & 1)
  {
    accel_z_axis += 2;
  }
  if((outbuf[5] >> 7) & 1)
  {
    accel_z_axis += 1;
  }
}

char WiiSensors::nunchuck_decode_byte(char x)
{
  x = (x ^ 0x17) + 0x17;
  return x;
}

//see http://wiibrew.org/wiki/Wiimote/Extension_Controllers#Wii_Motion_Plus
